from typing import Generator
from usefulgen.modules.basic import *
import itertools

def Wordlist(charset:str ,length:int, *, strict:bool=False, show_amount_combinations:bool=False) -> Generator[str, None, None]:
    """
    Generates a huge amount of strings combinations of a given length using a given charset, will use only the chars from the charset (if strict=True) or their type (if strict=False)
    \n
    Usage example:
    \n
    Wordlist("H",3, strict=False) --> First combination: AAA | Last combination: ZZZ | Other combination: OAL
    \n
    Wordlist("Hello",4, strict=True) --> First combination: HHHH | Last combination: oooo | Other combination: ollH
    
    :param charset: String charset that will be used to generate words with the same pattern. **It is case sensitive**.
    :type charset: str, list of different types of chars when strict=False -> a, A, 1, !, é
    :param length: Integer that will be used to generate words of this length.
    :type length: int
    :param strict: If True, only the character present in the charset will be used. If False, it will use the same kind of chars as the charset. **False by default**
    :type strict: bool, optional, default=False
    :param show_amount_combinations: If True, will print out the amount of combinations that will be generated in the process. **False by default**
    :type show_amount_combinations: bool, optional, default=False
    :raises TypeError: If charset or length is empty
    :raises TypeError: If length is given but is not an integer
    :raises ValueError: If length is given but is not higher than 0
    :raises TypeError: If charset is given but is not a string
    :raises TypeError: If strict is not empty but is not a boolean
    :raises TypeError: If show_amount_combinations is not empty but is not a boolean
    :return: A generator yielding words depending on your parameters
    :rtype: Generator[str]
    """
    if charset is None or length is None:
        raise TypeError(f"Both parameters must be filled (charset and length)")
    elif not isinstance(length, int):
        raise TypeError(f"Expected integer, received {length}")
    elif length <= 0:
        raise ValueError(f"length ({length}) must be higher than 0")
    elif not isinstance(charset, str):
        raise TypeError(f"Expected string, received {charset}")
    elif strict is not None and not isinstance(strict, bool):
        raise TypeError(f"Expected boolean type as strict, received \'{strict}\'")
    elif show_amount_combinations is not None and not isinstance(show_amount_combinations, bool):
        raise TypeError(f"Expected boolean type as show_amount_combinations, received \'{show_amount_combinations}\'")
    
    if strict: #checks for double entries
        chars:list = []
        for letter in charset:
            if letter not in chars:
                chars.append(letter)
        charset = str(''.join(chars))
        del chars

    def __DefineCharset(string:str) -> list:
        local_charset:list = []

        generators = list(lowerLatinAlphabet()), list(upperLatinAlphabet()), list(symbols()), list(symbolsCountrySpecific()), list(str(i) for i in numbers(0,9))
        for char in string:
            for gen in generators:
                if char in gen and char not in local_charset:
                    for sample in gen:
                        if sample not in local_charset:
                            local_charset.append(sample)
                    break
                elif gen == generators[-1] and char not in local_charset:
                    raise ValueError(f"\"{char}\" is not a valid character")
        return local_charset

    lcharset:list = []
    if not strict:
        lcharset = list(__DefineCharset(string=charset))
    elif strict:
        lcharset = list(charset)

    if show_amount_combinations:
            print(f"Generating {len(lcharset)**length} combinations...")
    
    already_yield = []

    for combination in itertools.product(lcharset, repeat=length):
        if combination not in already_yield:
            already_yield.append(combination)
            yield ''.join(combination)


def WordlistWithKnownParts(charset:str ,length:int, prefix:str=None, suffix:str=None, *, strict:bool=False, show_amount_combinations:bool=False) -> Generator[str, None, None]:
    """
    Same as Wordlist() but adds a suffix or/and a prefix to the output
    \n
    Refer to Wordlist() function's docstring for a complete docstring
    \n
    Usage example:
    \n
    WordlistWithKnownParts("johndoe", 4, "123", "!", strict=True) --> First combination: "123jjjj!" | Last combination: "123eeee!"
    \n
    WordlistWithKnownParts("john.doe", 5, None, "@gmail.com", strict=True) --> First combination: "jjjjj@gmail.com" | Last combination: "eeeee@gmail.com"
    
    :param: Please refer to Wordlist() docstring for the other parameters
    :param prefix: A prefix that will be added to each output
    :type prefix: str, optional
    :param suffix: A suffix that will be added to each output
    :type suffix: str, optional
    :raises TypeError: If suffix and prefix parameters are both empty
    :raises TypeError: If suffix or prefix are not strings (if not empty)
    :return: A generator yielding words with prefixes or/and suffixes
    :rtype: Generator[str]
    """
    if suffix is None and prefix is None:
        raise TypeError(f"A suffix or a prefix must be given (or both)")
    if suffix is None:
        suffix = ""
    elif prefix is None:
        prefix = ""
    if not isinstance(suffix, str) and not isinstance(prefix, str):
        raise TypeError(f"Expected strings as suffix or prefix")

    for combination in Wordlist(charset=charset, length=length, strict=strict, show_amount_combinations=show_amount_combinations):
            yield str(prefix+combination+suffix)


from typing import Generator

def upperLatinAlphabet() -> Generator[str, None, None]:
    """Generates a list of all the uppercase letters of the latin alphabet

    :return: A generator yielding each letters of the latin alphabet
    :rtype: Generator[str]
    """
    for i in range(65,91):
        yield chr(i)

def lowerLatinAlphabet() -> Generator[str, None, None]:
    """Generates a list of all the lowercase letters of the latin alphabet

    :return: A generator yielding each letters of the latin alphabet
    :rtype: Generator[str]
    """
    for i in range(97,123):
        yield chr(i)

def numbers(start:int, end:int) -> Generator[int, None, None]:
    """Generates a list of integers from 'start' to 'end' included without any leading zeros

    :param start: First integer of the list (inclusive)
    :type start: int
    :param end: Last integer of the list (inclusive)
    :type end: int
    :return: A generator yielding integers from 'start' to 'end' (inclusive)
    :rtype: Generator[int]
    :raises TypeError: If 'start' or 'end' is not an integer
    :raises ValueError: If 'start' is bigger than 'end'
    :raises ValueError: If 'start' or 'end' is not positive
    """
    if not isinstance(start, int) or not isinstance(end, int):
        raise TypeError(f"Both start ({start}) and end ({end}) must be integers")
    elif start >= end:
        raise ValueError(f"start ({start}) must be lower than end ({end})")
    elif start < 0 or end < 0:
        raise ValueError(f"Both start ({start}) and end ({end}) must be positive integers")
    
    for i in range(start,end+1):
        yield i

def numbersLeadingZeros(start:int, end:int, leading_zeros:int) -> Generator[str, None, None]:
    """Generates a list of strings of the integers from 'start' to 'end' included with the leading zeros
    
    :param start: First number of the list (inclusive)
    :type start: int
    :param end: Last integer of the list (inclusive)
    :type end: int
    :param leading_zeros: The number of leading zeros to add to a numerical sequence from right to left. For example, if leading_zeros=1, the sequence would be 1, 2, 3, ... If leading_zeros=2, the sequence would be 01, 02, 03, ...
    :type leading_zeros: int
    :return: A generator yielding strings of integers from 'start' to 'end' (inclusive)
    :rtype: Generator[str]
    :raises TypeError: If 'start', 'end' or 'leading_zeros' is not an integer
    :raises ValueError: If 'start' is bigger than 'end'
    :raises ValueError: If 'start', 'end' or 'leading_zeros' is not positive
    """
    if not isinstance(start, int) or not isinstance(end, int) or not isinstance(leading_zeros, int):
        raise TypeError(f"Both start ({start}), end ({end}) and leading_zeros ({leading_zeros}) must be integers")
    elif start >= end:
        raise ValueError(f"start ({start}) must be lower than end ({end})")
    elif start < 0 or end < 0 or leading_zeros < 0:
        raise ValueError(f"start ({start}), end ({end}) and leading_zeros ({leading_zeros}) must be positive integers")

    for i in range(start,end+1):
        i = str(i)
        while len(i) < leading_zeros:
            i = "0"+i
        yield i

def symbols(space:bool=True) -> Generator[str, None, None]:
    """Generates a list of all the most common symbols

    :param space: Determines if the space character (ASCII:32) will be yield or not
    :type space: bool, optional
    :return: A generator yielding each common symbols
    :rtype: Generator[str]
    :raises TypeError: If 'space' is not a boolean type
    """
    if not isinstance(space, bool):
        raise TypeError(f"space parameter ({space}) must be a bool-type (True/False)")

    if space:
        yield chr(32)
    for i in range(33,48):
        yield chr(i)
    for i in range(58,65):
        yield chr(i)
    for i in range(91,97):
        yield chr(i)
    for i in range(123,126):
        yield chr(i)

def symbolsCountrySpecific(country:str=None) -> Generator[str, None, None]:
    """Generates a list of all the letters considered as special characters from every country

    :param country: Determines the language from where the accents will be taken
    :type country: str, optional
    :return: A generator yielding each letters containing a special accents
    :rtype: Generator[str]
    :raises TypeError: If 'country' is not a string
    :raises ValueError: If 'country' is not supported (not in the countries list)
    """
    if country is not None:
        countries:dict = {
                "france":[192,194,196,198,199,200,201,202,203,206,207,212,217,219,220,224,226,228,230,231,232,233,234,235,238,239,244,249,251,252,338,339],
                "germany":[196,214,220,223,228,246,252],
                "italy":[192,200,201,204,210,217,224,232,233,236,242,249],
                "spain":[161,191,193,201,205,209,211,218,220,225,233,237,241,243,250,252]
                }

        if not isinstance(country, str):
            raise TypeError(f"country ({country}) is not a string type")
        elif str.lower(country) not in list(countries.keys()):
            raise ValueError(f"The country you entered ({country}) is not supported")
        
        for i in countries[str.lower(country)]:
            yield chr(i)

    else:
        yield chr(161)
        for i in range(191,215):
            yield chr(i)
        for i in range(216,222):
            yield chr(i)
        for i in range(223,247):
            yield chr(i)
        for i in range(248,253):
            yield chr(i)
        for i in range(338,340):
            yield chr(i)

def PiDecimals(start:int, end:int) -> Generator[int, None, None]:
    """Generates a list of a determined quantity of numbers after Pi's decimal
    
    :param start: Determines the start of the list after Pi's decimal (inclusive)
    :type start: int
    :param end: Determines the amount of numbers after Pi's decimal (inclusive)
    :type end: int
    :raises TypeError: If 'end' or 'start' is not int-type
    :raises ValueError: If 'end' or 'start' is < 1
    :raises ValueError: If 'start' is bigger than 'end'
    :return: A generator yielding numbers after Pi's decimal
    :rtype: Generator[int]
    """
    if not isinstance(start, int) or not isinstance(end, int):
        raise TypeError(f"Both start ({start}) and end ({end}) must be integers")
    elif start > end:
        raise ValueError(f"start ({start}) must be lower or equal than end ({end})")
    elif start < 1 or end < 1:
        raise ValueError(f"Both start ({start}) and end ({end}) must be >= 1")
    
    try:
        from decimal import Decimal, getcontext
    except ModuleNotFoundError or ImportError:
        raise ImportError("'decimal' module could not be imported. Since it is a built-in module, check your Python version!")

    validity = 5 # precalculates after x terms to insure the validity of the result
    
    def calc_pi(n:int) -> float:
        getcontext().prec = n+1  # Set the decimal precision to n+1
        pi = Decimal(0)
        for k in range(n):
            pi += (Decimal(1)/(16**k)) * (
                (Decimal(4)/(8*k+1)) -
                (Decimal(2)/(8*k+4)) -
                (Decimal(1)/(8*k+5)) -
                (Decimal(1)/(8*k+6))
            )
        return pi
    
    pi = str(calc_pi(end+validity))[2+start-1:-validity]
    for number in list(pi):
        yield int(number)


from setuptools import setup, find_packages

setup(
    name='usefulgen',
    version='1.0.3.1',
    packages=find_packages(),
    description='Python Module dedicated to time-saving generators, made by M4elstr0m',
    install_requires=['typing'],
    author='M4elstr0m',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/M4elstr0m/UsefulGenPy',
    download_url='https://pypi.org/project/usefulgen/',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License', 
    ],
)

#python setup.py sdist bdist_wheel
#twine upload dist/*